import json
from datetime import datetime, timedelta
from unittest.mock import patch
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.urls import reverse
from django.utils import timezone
from rest_framework.authtoken.models import Token
from swapper import load_model

from openwisp_controller.config.tests.utils import CreateDeviceGroupMixin
from openwisp_controller.geo.tests.utils import TestGeoMixin
from openwisp_users.tests.test_api import AuthenticationMixin
from openwisp_users.tests.utils import TestMultitenantAdminMixin
from openwisp_utils.tests import capture_any_output, catch_signal

from ... import settings as monitoring_settings
from ...monitoring.signals import post_metric_write, pre_metric_write
from ..api.serializers import WifiSessionSerializer
from ..signals import device_metrics_received
from . import DeviceMonitoringTestCase, TestWifiClientSessionMixin

start_time = timezone.now()
User = get_user_model()
Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
DeviceData = load_model('device_monitoring', 'DeviceData')
# needed for config.geo
Device = load_model('config', 'Device')
DeviceLocation = load_model('geo', 'DeviceLocation')
FloorPlan = load_model('geo', 'FloorPlan')
Location = load_model('geo', 'Location')
WifiClient = load_model('device_monitoring', 'WifiClient')
WifiSession = load_model('device_monitoring', 'WifiSession')
Group = load_model('openwisp_users', 'Group')


class TestDeviceApi(AuthenticationMixin, TestGeoMixin, DeviceMonitoringTestCase):
    """Tests API (device metric collection)."""

    location_model = Location
    object_location_model = DeviceLocation
    object_model = Device
    floorplan_model = FloorPlan
    # Exclude general metrics from the query
    metric_queryset = Metric.objects.exclude(object_id=None)
    # Exclude general charts from the query
    chart_queryset = Chart.objects.exclude(metric__object_id=None)
    _RESPONSE_KEYS = [
        'id',
        'name',
        'organization',
        'group',
        'mac_address',
        'key',
        'last_ip',
        'management_ip',
        'model',
        'os',
        'system',
        'notes',
        'config',
        'monitoring',
        'created',
        'modified',
        'charts',
        'x',
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Populate the ContentType cache to avoid queries during test
        ContentType.objects.get_for_model(Device)

    def setUp(self):
        self._login_admin()

    def tearDown(self):
        super().tearDown()
        cache.clear()

    def _login_admin(self):
        u = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(u)

    def _assert_device_info(self, device=None, data=None):
        self.assertEqual(data['id'], str(device.pk))
        self.assertEqual(data['name'], device.name)
        self.assertEqual(data['organization'], device.organization.pk)
        self.assertEqual(data['group'], device.group)
        self.assertEqual(data['mac_address'], device.mac_address)
        self.assertEqual(data['key'], device.key)
        self.assertEqual(data['last_ip'], device.last_ip)
        self.assertEqual(data['management_ip'], device.management_ip)
        self.assertEqual(data['model'], device.model)
        self.assertEqual(data['os'], device.os)
        self.assertEqual(data['system'], device.system)
        self.assertEqual(data['notes'], device.notes)
        self.assertIsNone(data['config'])

    def _assert_device_metrics_info(self, data=None, detail=True, charts=True):
        self.assertIn('monitoring', data)
        self.assertIn('status', data['monitoring'])
        if charts:
            self.assertEqual(len(list(data['charts'])), 7)
        if detail:
            self.assertIn('related_metrics', data['monitoring'])
            metrics = list(data['monitoring']['related_metrics'])
            self.assertEqual(metrics[0]['name'], 'CPU usage')
            self.assertEqual(metrics[0]['is_healthy'], True)
            self.assertEqual(metrics[1]['name'], 'Disk usage')
            self.assertEqual(metrics[1]['is_healthy'], True)
            self.assertEqual(metrics[2]['name'], 'Memory usage')
            self.assertEqual(metrics[2]['is_healthy'], True)
            self.assertEqual(metrics[3]['name'], 'wlan0 traffic')
            self.assertEqual(metrics[3]['is_healthy'], None)
            self.assertEqual(metrics[4]['name'], 'wlan0 wifi clients')
            self.assertEqual(metrics[4]['is_healthy'], None)
            self.assertEqual(metrics[5]['name'], 'wlan1 traffic')
            self.assertEqual(metrics[5]['is_healthy'], None)
            self.assertEqual(metrics[6]['name'], 'wlan1 wifi clients')
            self.assertEqual(metrics[6]['is_healthy'], None)

    def test_404(self):
        r = self._post_data(self.device_model().pk, '123', self._data())
        self.assertEqual(r.status_code, 404)

    def test_403(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        r = self.client.post(self._url(d.pk))
        self.assertEqual(r.status_code, 403)
        r = self._post_data(d.id, 'WRONG', self._data())
        self.assertEqual(r.status_code, 403)

    def test_400(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        r = self._post_data(d.id, d.key, {'interfaces': []})
        self.assertEqual(r.status_code, 400)
        r = self._post_data(d.id, d.key, {'interfaces': []})
        self.assertEqual(r.status_code, 400)
        r = self._post_data(
            d.id, d.key, {'type': 'DeviceMonitoring'}, time='23-08-2021 06:25:45'
        )
        self.assertEqual(r.status_code, 400)
        r = self._post_data(
            d.id, d.key, {'type': 'DeviceMonitoring', 'interfaces': [{}]}
        )
        self.assertEqual(r.status_code, 400)

    def test_200_none(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        data = {'type': 'DeviceMonitoring', 'interfaces': []}
        with self.assertNumQueries(4):
            r = self._post_data(d.id, d.key, data)
        self.assertEqual(r.status_code, 200)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 0)
        self.assertEqual(self.chart_queryset.count(), 0)
        data = {'type': 'DeviceMonitoring'}
        with self.assertNumQueries(2):
            r = self._post_data(d.id, d.key, data)
        self.assertEqual(r.status_code, 200)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 0)
        self.assertEqual(self.chart_queryset.count(), 0)
        d.delete(check_deactivated=False)
        r = self._post_data(d.id, d.key, data)
        self.assertEqual(r.status_code, 404)

    def test_200_create(self):
        self.create_test_data(no_resources=True)

    @patch('openwisp_monitoring.device.tasks.write_device_metrics.delay')
    def test_background_write(self, mocked_task):
        device = self._create_device(organization=self._create_org())
        data = self._data()
        r = self._post_data(device.id, device.key, data)
        self.assertEqual(r.status_code, 200)
        mocked_task.assert_called_once()

    def test_200_traffic_counter_incremented(self):
        dd = self.create_test_data(no_resources=True)
        d = self.device_model.objects.first()
        data2 = self._data()
        # creation of resources metrics can be avoided here as it is not involved
        # this speeds up the test by reducing requests made
        del data2['resources']
        data2['interfaces'][0]['statistics']['rx_bytes'] = 983
        data2['interfaces'][0]['statistics']['tx_bytes'] = 1567
        data2['interfaces'][1]['statistics']['rx_bytes'] = 2983
        data2['interfaces'][1]['statistics']['tx_bytes'] = 4567
        r = self._post_data(d.id, d.key, data2)
        self.assertEqual(r.status_code, 200)
        self.assertDataDict(dd.data, data2)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 4)
        self.assertEqual(self.chart_queryset.count(), 4)
        if_dict = {'wlan0': data2['interfaces'][0], 'wlan1': data2['interfaces'][1]}
        for ifname in ['wlan0', 'wlan1']:
            iface = if_dict[ifname]
            m = self.metric_queryset.get(name=f'{ifname} traffic', object_id=d.pk)
            points = self._read_metric(
                m, limit=10, order='-time', extra_fields=['tx_bytes']
            )
            self.assertEqual(len(points), 2)
            for field in ['rx_bytes', 'tx_bytes']:
                expected = iface['statistics'][field] - points[1][field]
                self.assertEqual(points[0][field], expected)
            m = self.metric_queryset.get(name=f'{ifname} wifi clients', object_id=d.pk)
            points = self._read_metric(m, limit=10, order='-time')
            self.assertEqual(len(points), len(iface['wireless']['clients']) * 2)

    def test_200_traffic_counter_reset(self):
        dd = self.create_test_data(no_resources=True)
        d = self.device_model.objects.first()
        data2 = self._data()
        # creation of resources metrics can be avoided here as it is not involved
        # this speeds up the test by reducing requests made
        del data2['resources']
        data2['interfaces'][0]['statistics']['rx_bytes'] = 50
        data2['interfaces'][0]['statistics']['tx_bytes'] = 20
        data2['interfaces'][1]['statistics']['rx_bytes'] = 80
        data2['interfaces'][1]['statistics']['tx_bytes'] = 120
        r = self._post_data(d.id, d.key, data2)
        self.assertEqual(r.status_code, 200)
        self.assertDataDict(dd.data, data2)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 4)
        self.assertEqual(self.chart_queryset.count(), 4)
        if_dict = {'wlan0': data2['interfaces'][0], 'wlan1': data2['interfaces'][1]}
        for ifname in ['wlan0', 'wlan1']:
            iface = if_dict[ifname]
            m = self.metric_queryset.get(name=f'{ifname} traffic', object_id=d.pk)
            points = self._read_metric(
                m, limit=10, order='-time', extra_fields=['tx_bytes']
            )
            self.assertEqual(len(points), 2)
            for field in ['rx_bytes', 'tx_bytes']:
                expected = iface['statistics'][field]
                self.assertEqual(points[0][field], expected)
            m = self.metric_queryset.get(name=f'{ifname} wifi clients', object_id=d.pk)
            points = self._read_metric(m, limit=10, order='-time')
            self.assertEqual(len(points), len(iface['wireless']['clients']) * 2)

    def test_device_with_location(self):
        self.create_test_data(no_resources=True)
        device = self.device_model.objects.first()
        location = self._create_location(
            organization=device.organization, type='indoor'
        )
        floorplan = self._create_floorplan(location=location)
        self._create_object_location(
            content_object=device, location=location, floorplan=floorplan
        )
        data2 = self._data()
        # creation of resources metrics can be avoided here as it is not involved
        # this speeds up the test by reducing requests made
        del data2['resources']
        additional_queries = 0 if self._is_timeseries_udp_writes else 1
        with self.assertNumQueries(21 + additional_queries):
            response = self._post_data(device.id, device.key, data2)
        # Ensure cache is working
        with self.assertNumQueries(13 + additional_queries):
            response = self._post_data(device.id, device.key, data2)
        self.assertEqual(response.status_code, 200)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 4)
        self.assertEqual(self.chart_queryset.count(), 4)
        for metric in self.metric_queryset.filter(object_id__isnull=False):
            points = self._read_metric(
                metric,
                limit=10,
                order='-time',
                extra_fields=['location_id', 'floorplan_id'],
            )
            for point in points:
                self.assertEqual(point['location_id'], str(location.id))
                self.assertEqual(point['floorplan_id'], str(floorplan.id))

    def test_200_multiple_measurements(self):
        dd = self._create_multiple_measurements(no_resources=True)
        # Add 1 for general metric and chart
        self.assertEqual(self.metric_queryset.count(), 4)
        self.assertEqual(self.chart_queryset.count(), 4)
        expected = {
            'wlan0': {'rx_bytes': 10000, 'tx_bytes': 6000},
            'wlan1': {'rx_bytes': 4587, 'tx_bytes': 2993},
        }
        # wlan0 traffic
        m = self.metric_queryset.get(name='wlan0 traffic', object_id=dd.pk)
        points = self._read_metric(
            m, limit=10, order='-time', extra_fields=['tx_bytes']
        )
        self.assertEqual(len(points), 4)
        expected = [700000000, 100000000, 399999676, 324]
        for i, point in enumerate(points):
            self.assertEqual(point['rx_bytes'], expected[i])
        expected = [300000000, 200000000, 99999855, 145]
        for i, point in enumerate(points):
            self.assertEqual(point['tx_bytes'], expected[i])
        c = m.chart_set.first()
        data = self._read_chart(c)
        # expected download wlan0
        self.assertEqual(data['traces'][0][1][-1], 1.2)
        # expected upload wlan0
        self.assertEqual(data['traces'][1][1][-1], 0.6)
        # wlan1 traffic
        m = self.metric_queryset.get(name='wlan1 traffic', object_id=dd.pk)
        points = self._read_metric(
            m, limit=10, order='-time', extra_fields=['tx_bytes']
        )
        self.assertEqual(len(points), 4)
        expected = [1000000000, 0, 1999997725, 2275]
        for i, point in enumerate(points):
            self.assertEqual(point['rx_bytes'], expected[i])
        expected = [500000000, 0, 999999174, 826]
        for i, point in enumerate(points):
            self.assertEqual(point['tx_bytes'], expected[i])
        c = m.chart_set.first()
        data = self._read_chart(c)
        # expected download wlan1
        self.assertEqual(data['traces'][0][1][-1], 3.0)
        # expected upload wlan1
        self.assertEqual(data['traces'][1][1][-1], 1.5)

    def test_200_no_date_supplied(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        data = self._data()
        netjson = json.dumps(data)
        url = self._url(d.id, d.key)
        r = self.client.post(url, netjson, content_type='application/json')
        self.assertEqual(r.status_code, 200)

    def test_404_disabled_organization(self):
        org = self._create_org(is_active=False)
        device = self._create_device(organization=org)
        with self.assertNumQueries(2):
            response = self._post_data(device.id, device.key, self._data())
        self.assertEqual(response.status_code, 404)
        self.assertEqual(self.metric_queryset.count(), 0)
        self.assertEqual(self.chart_queryset.count(), 0)

    def test_device_activate_deactivate(self):
        # "self.create_test_data" creates a device and makes
        # a POST request to DeviceMetricView ensuring that
        # the device is cached.
        self.create_test_data(no_resources=True)
        device = self.device_model.objects.first()
        data = {'type': 'DeviceMonitoring'}
        with self.assertNumQueries(2):
            response = self._post_data(device.id, device.key, data)

        # Deactivating the device will invalidate the cache.
        # The view will only allow readonly requests (GET).
        device.deactivate()
        response = self.client.get(self._url(device.pk, device.key))
        self.assertEqual(response.status_code, 200)
        with self.assertNumQueries(1):
            response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 404)

        # Re-activating the device will allow POST requests again.
        device.activate()
        with self.assertNumQueries(4):
            response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)

    def test_garbage_wireless_clients(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        garbage_interfaces = [
            {'name': 'garbage1', 'wireless': {'clients': {}}},
            {'name': 'garbage2', 'wireless': {'clients': [{'what?': 'mac missing'}]}},
            {'name': 'garbage3', 'wireless': {}},
        ]
        for garbage_interface in garbage_interfaces:
            interface = self._data()['interfaces'][0]
            interface.update(garbage_interface)
            r = self._post_data(
                d.id, d.key, {'type': 'DeviceMonitoring', 'interfaces': [interface]}
            )
            with self.subTest(garbage_interface):
                self.assertEqual(r.status_code, 400)

    @patch.object(monitoring_settings, 'AUTO_CHARTS', return_value=[])
    def test_auto_chart_disabled(self, *args):
        # Add 1 for general chart
        self.assertEqual(self.chart_queryset.count(), 0)
        o = self._create_org()
        d = self._create_device(organization=o)
        self._post_data(d.id, d.key, self._data())
        # Add 1 for general chart
        self.assertEqual(self.chart_queryset.count(), 0)

    def test_get_device_metrics_200(self):
        dd = self.create_test_data()
        d = self.device_model.objects.get(pk=dd.pk)
        with self.assertNumQueries(17):
            r = self.client.get(self._url(d.pk.hex, d.key))
        with self.assertNumQueries(16):
            r = self.client.get(self._url(d.pk.hex, d.key))
        self.assertEqual(r.status_code, 200)

        with self.subTest('Test device metrics 200 without the device key'):
            r1 = self.client.get(self._url(d.pk.hex))
            self.assertEqual(r1.status_code, 200)
            for key in self._RESPONSE_KEYS:
                self.assertIn(key, r.data.keys())

        with self.subTest('Test device information in the response'):
            self._assert_device_info(device=d, data=r.data)

        with self.subTest('Test device metrics in the response'):
            self._assert_device_metrics_info(data=r.data)

        with self.subTest('Test device charts in the response'):
            self.assertIn('x', r.data)
            charts = r.data['charts']
            for chart in charts:
                self.assertIn('traces', chart)
                self.assertIn('title', chart)
                self.assertIn('description', chart)
                self.assertIn('type', chart)
                self.assertIn('summary', chart)
                self.assertIsInstance(chart['summary'], dict)
                self.assertIn('summary_labels', chart)
                self.assertIsInstance(chart['summary_labels'], list)
                self.assertIn('unit', chart)
                self.assertIn('colors', chart)
                self.assertIn('colorscale', chart)
            # test charts order
            self.assertEqual(charts[0]['title'], 'WiFi clients: wlan0')
            self.assertEqual(charts[1]['title'], 'WiFi clients: wlan1')
            self.assertEqual(charts[2]['title'], 'Traffic: wlan0')
            self.assertEqual(charts[3]['title'], 'Traffic: wlan1')
            self.assertEqual(charts[4]['title'], 'Memory Usage')
            self.assertEqual(charts[5]['title'], 'CPU Load')
            self.assertEqual(charts[6]['title'], 'Disk Usage')

    def test_get_device_metrics_list_200(self):
        dd1 = self.create_test_data()
        url = reverse('monitoring:api_monitoring_device_list')
        d1 = self.device_model.objects.get(pk=dd1.pk)
        o2 = self._create_org(name='test org 2')
        d2 = self._create_device(
            name='test-device-2', mac_address='00:11:22:33:44:66', organization=o2
        )
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.data['count'], 2)
        self.assertEqual(r.data['next'], None)
        self.assertEqual(r.data['previous'], None)
        results = r.data['results'][0].keys()
        _RESPONSE_KEYS = self._RESPONSE_KEYS.copy()
        _RESPONSE_KEYS.remove('x')
        _RESPONSE_KEYS.remove('charts')
        for key in _RESPONSE_KEYS:
            self.assertIn(key, results)
        self._assert_device_info(device=d1, data=r.data['results'][0])
        self._assert_device_info(device=d2, data=r.data['results'][1])
        self._assert_device_metrics_info(
            data=r.data['results'][0], detail=False, charts=False
        )

        with self.subTest('Test filtering using monitoring health status'):
            r = self.client.get(f'{url}?monitoring__status=ok')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d1, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )
            # Update device (d2) health status to 'critical'
            d2.monitoring.update_status('critical')
            r = self.client.get(f'{url}?monitoring__status=critical')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d2, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )

        with self.subTest('Test filtering using organization id'):
            r = self.client.get(f'{url}?organization={d1.organization.id}')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d1, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )
            r = self.client.get(f'{url}?organization={d2.organization.id}')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d2, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )

        with self.subTest('Test filtering using organization slug'):
            r = self.client.get(f'{url}?organization_slug={d1.organization.slug}')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d1, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )
            r = self.client.get(f'{url}?organization_slug={d2.organization.slug}')
            self.assertEqual(r.data['count'], 1)
            self._assert_device_info(device=d2, data=r.data['results'][0])
            self._assert_device_metrics_info(
                data=r.data['results'][0], detail=False, charts=False
            )

    def test_get_device_metrics_histogram_ignore_x(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        m = self._create_object_metric(content_object=d, name='applications')
        self._create_chart(metric=m, configuration='histogram')
        self._create_multiple_measurements(create=False, no_resources=True, count=2)
        r = self.client.get(self._url(d.pk.hex, d.key))
        self.assertEqual(r.status_code, 200)
        self.assertTrue(len(r.data['x']) > 50)

    def test_get_device_metrics_1d(self):
        dd = self.create_test_data()
        d = self.device_model.objects.get(pk=dd.pk)
        r = self.client.get('{0}&time=1d'.format(self._url(d.pk, d.key)))
        self.assertEqual(r.status_code, 200)
        self.assertIsInstance(r.data['charts'], list)

    def test_get_device_metrics_404(self):
        r = self.client.get(self._url('WRONG', 'MADEUP'))
        self.assertEqual(r.status_code, 404)

    def test_get_device_metrics_401(self):
        d = self._create_device(organization=self._create_org())
        self.client.logout()
        # try to access the device metrics
        # ie. api view without authentication
        r = self.client.get(self._url(d.pk))
        self.assertEqual(r.status_code, 401)

    def test_get_device_list_metrics_401(self):
        self._create_device(organization=self._create_org())
        self.client.logout()
        # try to access the device metrics list
        # ie. api view without authentication
        url = reverse('monitoring:api_monitoring_device_list')
        r = self.client.get(url)
        self.assertEqual(r.status_code, 401)

    def test_get_device_metrics_400_bad_time_range(self):
        d = self._create_device(organization=self._create_org())
        url = '{0}&time=3w'.format(self._url(d.pk, d.key))
        r = self.client.get(url)
        self.assertEqual(r.status_code, 400)

    def test_get_device_metrics_csv(self):
        d = self._create_device(organization=self._create_org())
        self._create_multiple_measurements(create=False, count=2)
        m = self._create_object_metric(
            content_object=d, name='applications', configuration='get_top_fields'
        )
        self._create_chart(metric=m, configuration='histogram')
        m.write(None, extra_values={'http2': 90, 'ssh': 100, 'udp': 80, 'spdy': 70})
        r = self.client.get('{0}&csv=1'.format(self._url(d.pk, d.key)))
        self.assertEqual(r.get('Content-Disposition'), 'attachment; filename=data.csv')
        self.assertEqual(r.get('Content-Type'), 'text/csv')
        rows = r.content.decode('utf8').strip().split('\n')
        header = rows[0].strip().split(',')
        self.assertEqual(
            header,
            [
                'time',
                'wifi_clients - WiFi clients: wlan0',
                'wifi_clients - WiFi clients: wlan1',
                'download - Traffic: wlan0',
                'upload - Traffic: wlan0',
                'download - Traffic: wlan1',
                'upload - Traffic: wlan1',
                'memory_usage - Memory Usage',
                'CPU_load - CPU Load',
                'disk_usage - Disk Usage',
            ],
        )
        last_line_before_histogram = rows[-5].strip().split(',')
        self.assertEqual(
            last_line_before_histogram,
            [
                last_line_before_histogram[0],
                '1',
                '2',
                '0.4',
                '0.1',
                '2.0',
                '1.0',
                '9.73',
                '0.0',
                '8.27',
            ],
        )
        self.assertEqual(rows[-4].strip(), '')
        self.assertEqual(rows[-3].strip(), 'Histogram')
        self.assertEqual(rows[-2].strip().split(','), ['ssh', '100.0'])
        self.assertEqual(rows[-1].strip().split(','), ['http2', '90.0'])

    def test_histogram_csv_none_value(self):
        d = self._create_device(organization=self._create_org())
        m = self._create_object_metric(content_object=d, name='applications')
        mock = {
            'traces': [('http2', [100]), ('ssh', [None])],
            'x': ['2020-06-15 00:00'],
            'summary': {'http2': 100, 'ssh': None},
        }
        c = self._create_chart(metric=m, configuration='histogram')
        with patch(
            'openwisp_monitoring.monitoring.base.models.AbstractChart.read',
            return_value=mock,
        ):
            self.assertEqual(self._read_chart(c), mock)
            r = self.client.get('{0}&csv=1'.format(self._url(d.pk, d.key)))
        self.assertEqual(r.get('Content-Disposition'), 'attachment; filename=data.csv')
        self.assertEqual(r.get('Content-Type'), 'text/csv')
        rows = r.content.decode('utf8').strip().split('\n')
        self.assertEqual(rows[-2].strip().split(','), ['http2', '100'])
        self.assertEqual(rows[-1].strip().split(','), ['ssh', '0'])

    def test_get_device_metrics_empty(self):
        d = self._create_device(organization=self._create_org())
        m = self._create_object_metric(name='test_metric', content_object=d)
        c = Chart(metric=m, configuration='dummy')  # empty chart
        c.full_clean()
        c.save()
        # Add 1 for general chart
        self.assertEqual(self.chart_queryset.count(), 1)
        response = self.client.get(self._url(d.pk.hex, d.key))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['charts'], [])

    def test_get_device_metrics_400_bad_timezone(self):
        dd = self.create_test_data(no_resources=True)
        d = self.device_model.objects.get(pk=dd.pk)
        wrong_timezone_values = (
            'wrong',
            'America/Lima%27);%20DROP%20DATABASE%20test2;',
            'Europe/Cazzuoli',
        )
        for tz_value in wrong_timezone_values:
            url = '{0}&timezone={1}'.format(self._url(d.pk, d.key), tz_value)
            r = self.client.get(url)
            self.assertEqual(r.status_code, 400)
            self.assertIn('Unkown Time Zone', r.data)

    def test_device_metrics_received_signal(self):
        d = self._create_device(organization=self._create_org())
        dd = DeviceData(name='test-device', pk=d.pk)
        data = self._data()
        self._create_object_metric(name='ping', configuration='ping', content_object=d)
        time = start_time.strftime('%d-%m-%Y_%H:%M:%S.%f')
        with catch_signal(device_metrics_received) as handler:
            response = self._post_data(d.id, d.key, data, time=time)
        request = response.renderer_context['request']
        handler.assert_called_once_with(
            instance=dd,
            request=request,
            sender=DeviceData,
            signal=device_metrics_received,
            time=start_time,
            current=False,
        )

    @capture_any_output()
    def test_invalid_chart_config(self):
        # Tests if chart_config is invalid, then it is skipped and API does not fail
        d = self._create_device(organization=self._create_org())
        m = self._create_object_metric(name='test_metric', content_object=d)
        c = self._create_chart(metric=m, test_data=None)
        c.configuration = 'invalid'
        c.save()
        response = self.client.get(self._url(d.pk.hex, d.key))
        self.assertEqual(response.status_code, 200)

    def test_available_memory(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        data = self._data()
        data['resources']['memory']['free'] = 224497664
        with self.subTest('Test without available memory'):
            del data['resources']['memory']['available']
            r = self._post_data(d.id, d.key, data)
            m = self.metric_queryset.get(key='memory')
            metric_data = self._read_metric(m, order='-time', extra_fields='*')[0]
            self.assertEqual(metric_data['percent_used'], 9.729419481797308)
            self.assertIsNone(metric_data.get('available_memory'))
            self.assertEqual(r.status_code, 200)
        with self.subTest('Test when available memory is less than free memory'):
            data['resources']['memory']['available'] = 2232664
            r = self._post_data(d.id, d.key, data)
            metric_data = self._read_metric(m, order='-time', extra_fields='*')[0]
            self.assertEqual(metric_data['percent_used'], 9.729419481797308)
            self.assertEqual(metric_data['available_memory'], 2232664)
            self.assertEqual(r.status_code, 200)
        with self.subTest('Test when available memory is greater than free memory'):
            data['resources']['memory']['available'] = 225567664
            r = self._post_data(d.id, d.key, data)
            m = self.metric_queryset.get(key='memory')
            metric_data = self._read_metric(m, order='-time', extra_fields='*')[0]
            self.assertEqual(metric_data['percent_used'], 9.301032356920302)
            self.assertEqual(metric_data['available_memory'], 225567664)
            self.assertEqual(r.status_code, 200)

    def test_get_device_status_200(self):
        dd = self.create_test_data(no_resources=True)
        d = self.device_model.objects.get(pk=dd.pk)
        url = self._url(d.pk.hex, d.key)
        # status not requested
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)
        self.assertNotIn('status', r.data)
        # status requested
        r = self.client.get(f'{url}&status=1')
        self.assertEqual(r.status_code, 200)
        self.assertIn('data', r.data)
        self.assertIsInstance(r.data['data'], dict)
        self.assertEqual(dd.data, r.data['data'])

    def test_garbage_interface_properties(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        garbage_interfaces = [
            {'type': 1},
            {'uptime': 'string'},
            {'up': 'up'},
            {'mac': 1},
            {'mtu': 'string'},
            {'txqueuelen': 'string'},
            {'speed': 0},
            {'multicast': 1},
            {'type': 'bridge', 'bridge_members': [1, 2]},
            {'stp': 1},
        ]
        number = 1
        for garbage_interface in garbage_interfaces:
            interface = self._data()['interfaces'][0]
            interface.update(garbage_interface)
            interface['name'] = f'garbage{number}'
            r = self._post_data(
                d.id, d.key, {'type': 'DeviceMonitoring', 'interfaces': [interface]}
            )
            number += 1
            with self.subTest(garbage_interface):
                self.assertEqual(r.status_code, 400)

    def test_mobile_properties(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {
                            'lte': {'rsrp': -75, 'rsrq': -8, 'rssi': -51, 'snr': 13}
                        },
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        mobile_data = DeviceData(pk=device.pk).data['interfaces'][0]['mobile']
        with self.subTest('check mobile interface static properties'):
            self.assertEqual(mobile_data['connection_status'], 'connected')
            self.assertEqual(mobile_data['imei'], '300000001234567')
            self.assertEqual(mobile_data['model'], 'MC7430')
            self.assertEqual(mobile_data['operator_code'], '50502')
            self.assertEqual(mobile_data['operator_name'], 'YES OPTUS')
            self.assertEqual(mobile_data['power_status'], 'on')
        with self.subTest('ensure signal data is converted to float'):
            # ensure numbers are stored as floats,
            # lua cannot be forced to send floats so we need to force it
            self.assertIsInstance(mobile_data['signal']['lte']['rsrp'], float)
            self.assertIsInstance(mobile_data['signal']['lte']['rsrq'], float)
            self.assertIsInstance(mobile_data['signal']['lte']['rssi'], float)
            self.assertIsInstance(mobile_data['signal']['lte']['snr'], float)
            self.assertDictEqual(
                mobile_data['signal'],
                {'lte': {'rsrp': -75.00, 'rsrq': -8.00, 'rssi': -51.00, 'snr': 13.00}},
            )

    def test_empty_mobile_signal_data(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {'threshold': {'rssi': '0.0'}},
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        mobile_data = DeviceData(pk=device.pk).data['interfaces'][0]['mobile']
        with self.subTest('check mobile interface static properties'):
            self.assertEqual(mobile_data['imei'], '300000001234567')
            self.assertEqual(mobile_data['model'], 'MC7430')
            self.assertIn('signal', mobile_data)
            self.assertIn('threshold', mobile_data['signal'])

    def test_garbage_mobile_properties(self):
        o = self._create_org()
        d = self._create_device(organization=o)
        interface = {
            'name': 'mobile',
            'mac': '00:00:00:00:00:00',
            'mtu': 1900,
            'multicast': True,
            'txqueuelen': 1000,
            'type': 'modem-manager',
            'up': True,
        }
        garbage_data = [
            {'connection_status': 'connected'},
            {'imei': '300000001234567'},
            {
                'connection_status': 'connected',
                'imei': '300000001234567',
                'manufacturer': 'Sierra Wireless, Incorporated',
                'model': 'MC7430',
                'operator_code': '50502',
                'operator_name': 'YES OPTUS',
                'power_status': 'on',
                'signal': {'lte': {'rsrp': -75}},
            },
            {
                'connection_status': 'connected',
                'imei': '300000001234567',
                'manufacturer': 'Sierra Wireless, Incorporated',
                'model': 'MC7430',
                'operator_code': '50502',
                'operator_name': 'YES OPTUS',
                'power_status': 'on',
                'signal': {
                    'lte': {'rsrp': '-75', 'rsrq': '-8', 'rssi': '-51', 'snr': '13'}
                },
            },
        ]
        number = 0
        for mobile_data in garbage_data:
            interface_data = interface.copy()
            interface_data['mobile'] = mobile_data
            interface_data['name'] += str(number)
            r = self._post_data(
                d.id,
                d.key,
                {'type': 'DeviceMonitoring', 'interfaces': [interface_data]},
            )
            number += 1
            with self.subTest(interface_data['name']):
                self.assertEqual(r.status_code, 400)

    def test_mobile_charts(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        charts_count = self.chart_queryset.count()
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {
                            'lte': {'rsrp': -75, 'rsrq': -8, 'rssi': -51, 'snr': 13},
                        },
                    },
                }
            ],
        }
        self._post_data(device.id, device.key, data)
        data['interfaces'][0]['mobile']['signal'].update(
            {'umts': {'ecio': 2, 'rscp': -14, 'rssi': -80}}
        )
        self._post_data(device.id, device.key, data)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(charts[0]['summary']['signal_strength'], -51.0)
        self.assertEqual(charts[0]['summary']['signal_power'], -75.0)
        self.assertEqual(charts[1]['summary']['signal_quality'], -8.0)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], 13.0)
        self.assertEqual(charts[2]['summary']['access_tech'], 4.0)
        # ensure correct color-coding
        self.assertEqual(
            charts[2]['colorscale']['map'],
            [
                [5, '#377873', '5g'],
                [4, '#67c368', 'lte'],
                [3, '#efdd50', 'umts'],
                [2, '#df7514', 'evdo'],
                [1, '#dd5817', 'cdma1x'],
                [0, '#b42a0c', 'gsm'],
            ],
        )
        self.assertEqual(self.chart_queryset.count(), charts_count + 3)

    def test_5g_charts(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {
                            'lte': {'rsrp': -75, 'rsrq': -8, 'rssi': -51, 'snr': 13},
                            '5g': {'rsrp': -70, 'rsrq': -7, 'snr': 12},
                        },
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(charts[0]['summary']['signal_power'], -70.0)
        self.assertEqual(charts[0]['summary']['signal_strength'], None)
        self.assertEqual(charts[1]['summary']['signal_quality'], -7.0)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], 12.0)
        self.assertEqual(charts[2]['summary']['access_tech'], 5)

    def test_umts_rscp_missing(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {'umts': {'ecio': -5, 'rssi': -69}},
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(len(charts), 3)
        self.assertEqual(charts[0]['summary']['signal_strength'], -69.0)
        self.assertEqual(charts[0]['summary']['signal_power'], None)
        self.assertEqual(charts[1]['summary']['signal_quality'], None)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], -5.0)
        self.assertEqual(charts[2]['summary']['access_tech'], 3.0)

    def test_umts_rssi_missing(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {'umts': {'ecio': -3.50, 'rscp': -96.00}},
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(len(charts), 3)
        self.assertEqual(charts[0]['summary']['signal_strength'], None)
        self.assertEqual(charts[0]['summary']['signal_power'], -96.00)
        self.assertEqual(charts[1]['summary']['signal_quality'], None)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], -4)
        self.assertEqual(charts[2]['summary']['access_tech'], 3.0)

    def test_cdma_charts(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {'cdma1x': {'ecio': -5, 'rssi': -69}},
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(len(charts), 3)
        self.assertEqual(charts[0]['summary']['signal_strength'], -69.0)
        self.assertEqual(charts[0]['summary']['signal_power'], None)
        self.assertEqual(charts[1]['summary']['signal_quality'], None)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], -5.0)
        self.assertEqual(charts[2]['summary']['access_tech'], 1.0)

    def test_evdo_charts(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {
                            'evdo': {'ecio': -5, 'rssi': -69, 'io': -70, 'sinr': -11},
                        },
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(len(charts), 3)
        self.assertEqual(charts[0]['summary']['signal_strength'], -69.0)
        self.assertEqual(charts[0]['summary']['signal_power'], None)
        self.assertEqual(charts[1]['summary']['signal_quality'], None)
        self.assertEqual(charts[1]['summary']['signal_to_noise_ratio'], -11.0)
        self.assertEqual(charts[2]['summary']['access_tech'], 2.0)

    def test_gsm_charts(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '300000001234567',
                        'manufacturer': 'Sierra Wireless, Incorporated',
                        'model': 'MC7430',
                        'operator_code': '50502',
                        'operator_name': 'YES OPTUS',
                        'power_status': 'on',
                        'signal': {'gsm': {'rssi': -70}},
                    },
                }
            ],
        }
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        charts = response.data['charts']
        self.assertEqual(len(charts), 2)
        self.assertEqual(charts[0]['summary']['signal_power'], None)
        self.assertEqual(charts[0]['summary']['signal_strength'], -70.0)

    def test_mobile_signal_missing(self):
        org = self._create_org()
        device = self._create_device(organization=org)
        data = {
            'type': 'DeviceMonitoring',
            'interfaces': [
                {
                    'name': 'mobile0',
                    'mac': '00:00:00:00:00:00',
                    'mtu': 1900,
                    'multicast': True,
                    'txqueuelen': 1000,
                    'type': 'modem-manager',
                    'up': True,
                    'mobile': {
                        'connection_status': 'connected',
                        'imei': '865847055230161',
                        'manufacturer': 'QUALCOMM INCORPORATED',
                        'model': 'QUECTEL Mobile Broadband Module',
                        'operator_code': '22250',
                        'operator_name': 'Iliad',
                        'power_status': 'on',
                        'signal': {},
                    },
                }
            ],
        }
        self._post_data(device.id, device.key, data)
        response = self.client.get(self._url(device.pk.hex, device.key))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['charts'], [])
        dd = DeviceData(name=device.name, pk=device.pk)
        self.assertEqual(
            dd.data['interfaces'][0]['mobile'], data['interfaces'][0]['mobile']
        )

        with self.subTest('signal key not present at all'):
            del data['interfaces'][0]['mobile']['signal']
            response = self._post_data(device.id, device.key, data)
            self.assertEqual(response.status_code, 200)

    def test_pre_metric_write_signal(self):
        d = self._create_device(organization=self._create_org())
        data = {'type': 'DeviceMonitoring', 'resources': {'cpus': 1, 'load': [0, 0, 0]}}
        self._create_object_metric(
            name='CPU usage', configuration='cpu', object_id=d.id
        )
        values = {'cpu_usage': 0.0, 'load_1': 0.0, 'load_5': 0.0, 'load_15': 0.0}
        time = start_time.strftime('%d-%m-%Y_%H:%M:%S.%f')
        with catch_signal(pre_metric_write) as handler:
            self._post_data(d.id, d.key, data, time=time)
        signal_calls = handler.call_args_list
        # assert signal is called once
        self.assertEqual(len(signal_calls), 1)
        signal_arguments = signal_calls[0][1]
        # remove metric from signal arguments
        del signal_arguments['metric']
        expected_arguments = dict(
            signal=pre_metric_write,
            sender=Metric,
            values=values,
            time=start_time,
            current=False,
        )
        self.assertEqual(signal_calls[0][1], expected_arguments)

    def test_post_metric_write_signal(self):
        d = self._create_device(organization=self._create_org())
        data = {'type': 'DeviceMonitoring', 'resources': {'cpus': 1, 'load': [0, 0, 0]}}
        self._create_object_metric(
            name='CPU usage', configuration='cpu', object_id=d.id
        )
        values = {'cpu_usage': 0.0, 'load_1': 0.0, 'load_5': 0.0, 'load_15': 0.0}
        time = start_time.strftime('%d-%m-%Y_%H:%M:%S.%f')
        with catch_signal(post_metric_write) as handler:
            self._post_data(d.id, d.key, data, time=time)
        signal_calls = handler.call_args_list
        # assert signal is called once
        self.assertEqual(len(signal_calls), 1)
        signal_arguments = signal_calls[0][1]
        # remove metric from signal arguments
        del signal_arguments['metric']
        expected_arguments = dict(
            signal=post_metric_write,
            sender=Metric,
            values=values,
            time=start_time.isoformat(),
            current=False,
        )
        self.assertEqual(signal_calls[0][1], expected_arguments)

    def test_device_custom_date_metrics(self):
        now = datetime.now()
        dd = self.create_test_data()
        d = self.device_model.objects.get(pk=dd.pk)

        def _assert_chart_group(url, status_code, expected):
            response = self.client.get(url)
            chart_group = Chart._get_group_map(expected[0]).items()
            self.assertEqual(response.status_code, status_code)
            self.assertIn(expected, chart_group)

        with self.subTest('Test custom grouping between 1 to 2 days'):
            custom_date_query = '&start=2022-10-02%2000:00:00&end=2022-10-04%2023:59:59'
            url = f'{self._url(d.pk, d.key)}{custom_date_query}'
            _assert_chart_group(url, 200, ('2d', '10m'))

        with self.subTest('Test custom grouping between 3 to 6 days'):
            custom_date_query = '&start=2022-10-02%2000:00:00&end=2022-10-06%2023:59:59'
            url = f'{self._url(d.pk, d.key)}{custom_date_query}'
            _assert_chart_group(url, 200, ('4d', '25m'))

        with self.subTest('Test custom grouping between 8 to 27 days'):
            custom_date_query = '&start=2022-09-29%2000:00:00&end=2022-10-11%2015:50:56'
            url = f'{self._url(d.pk, d.key)}{custom_date_query}'
            _assert_chart_group(url, 200, ('12d', '2h'))

        with self.subTest('Test custom grouping between 28 to 364 days'):
            custom_date_query = '&start=2022-07-04%2000:00:00&end=2022-10-11%2015:50:56'
            url = f'{self._url(d.pk, d.key)}{custom_date_query}'
            _assert_chart_group(url, 200, ('99d', '1d'))

        with self.subTest('Test invalid custom dates'):
            invalid_custom_dates = (
                '&start=2022-07-04&end=2022-10-11',
                '&start=September&end=October',
                '&start=2022-07-04&end=10-11-2022%2015:50:56',
                '&start=09-08-2022%2000:00:00&end=10-11-2022%2016:39:29',
            )
            for invalid_date in invalid_custom_dates:
                url = f'{self._url(d.pk, d.key)}{invalid_date}'
                r = self.client.get(url)
                self.assertEqual(r.status_code, 400)
                self.assertIn(
                    'Incorrect custom date format, should be YYYY-MM-DD H:M:S', r.data
                )
        with self.subTest(
            'Test device metrics when start date is greater than end date'
        ):
            start_greater_than_end = (
                '&start=2022-09-23%2000:00:00&end=2022-09-13%2023:59:59'
            )
            url = f'{self._url(d.pk, d.key)}{start_greater_than_end}'
            r = self.client.get(url)
            self.assertEqual(r.status_code, 400)
            self.assertIn(
                'start_date cannot be greater than end_date',
                r.data,
            )
        with self.subTest(
            'Test device metrics when start date is greater than today date'
        ):
            start_greater = (now + timedelta(days=5)).strftime('%Y-%m-%d')
            end_greater = (now + timedelta(days=9)).strftime('%Y-%m-%d')
            start_greater_than_now = (
                f'&start={start_greater}%2000:00:00&end={end_greater}%2023:59:59'
            )
            url = f'{self._url(d.pk, d.key)}{start_greater_than_now}'
            r = self.client.get(url)
            self.assertEqual(r.status_code, 400)
            self.assertIn(
                "start_date cannot be greater than today's date",
                r.data,
            )
        with self.subTest(
            'Test device metrics when end date is greater than today date'
        ):
            start_lesser = (now - timedelta(days=2)).strftime('%Y-%m-%d')
            end_greater = (now + timedelta(days=5)).strftime('%Y-%m-%d')
            end_greater_than_now = (
                f'&start={start_lesser}%2000:00:00&end={end_greater}%2023:59:59'
            )
            url = f'{self._url(d.pk, d.key)}{end_greater_than_now}'
            r = self.client.get(url)
            self.assertEqual(r.status_code, 400)
            self.assertIn(
                "end_date cannot be greater than today's date",
                r.data,
            )

        with self.subTest(
            'Test device metrics when date range is greater than 365 days'
        ):
            greater_than_365_days = (
                '&start=2021-11-11%2000:00:00&end=2022-11-12%2023:59:59'
            )
            url = f'{self._url(d.pk, d.key)}{greater_than_365_days}'
            r = self.client.get(url)
            self.assertEqual(r.status_code, 400)
            self.assertIn(
                "The date range shouldn't be greater than 365 days",
                r.data,
            )

        with self.subTest('Test device metrics csv export with custom dates'):
            end_date = now.strftime('%Y-%m-%d')
            start_date = (now - timedelta(days=2)).strftime('%Y-%m-%d')
            self._create_multiple_measurements(create=False, count=2)
            m = self._create_object_metric(
                content_object=d, name='applications', configuration='get_top_fields'
            )
            self._create_chart(metric=m, configuration='histogram')
            m.write(None, extra_values={'http2': 90, 'ssh': 100, 'udp': 80, 'spdy': 70})
            custom_date_query = (
                f'&start={start_date}%2000:00:00&end={end_date}%2000:00:00'
            )
            url = f'{self._url(d.pk, d.key)}{custom_date_query}&csv=1'
            response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(
                response.get('Content-Disposition'), 'attachment; filename=data.csv'
            )
            self.assertEqual(response.get('Content-Type'), 'text/csv')
            rows = response.content.decode('utf8').strip().split('\n')
            self.assertEqual(rows[-3].strip(), 'Histogram')
            self.assertEqual(rows[-2].strip().split(','), ['ssh', '100.0'])
            self.assertEqual(rows[-1].strip().split(','), ['http2', '90.0'])

    def test_missing_rx_bytes(self):
        """
        Regression test for:
        https://github.com/openwisp/openwisp-monitoring/issues/595
        """
        data = self._data()
        del data['interfaces'][0]['statistics']['rx_bytes']
        dd = self.create_test_data(no_resources=True, data=data, assertions=False)
        qs = Metric.objects.filter(
            object_id=dd.pk, key='traffic', main_tags={'ifname': 'wlan0'}
        )
        self.assertEqual(qs.count(), 1)
        m = qs.first()
        points = self._read_metric(m, limit=1, extra_fields=['tx_bytes'])
        self.assertEqual(points[0].get('tx_bytes'), 145)
        self.assertEqual(points[0].get('rx_bytes'), 0)

    def test_missing_tx_bytes(self):
        data = self._data()
        del data['interfaces'][0]['statistics']['tx_bytes']
        dd = self.create_test_data(no_resources=True, data=data, assertions=False)
        qs = Metric.objects.filter(
            object_id=dd.pk, key='traffic', main_tags={'ifname': 'wlan0'}
        )
        self.assertEqual(qs.count(), 1)
        m = qs.first()
        points = self._read_metric(m, limit=1, extra_fields=['tx_bytes'])
        self.assertEqual(points[0].get('rx_bytes'), 324)
        self.assertEqual(points[0].get('tx_bytes'), 0)


class TestGeoApi(TestGeoMixin, AuthenticationMixin, DeviceMonitoringTestCase):
    location_model = Location
    object_location_model = DeviceLocation
    object_model = Device
    floorplan_model = FloorPlan

    def _login_admin(self):
        User = get_user_model()
        admin = User.objects.create_superuser('admin', 'admin', 'test@test.com')
        self.client.force_login(admin)

    def test_api_location_geojson(self):
        device_location = self._create_object_location()
        device_location.device.monitoring.update_status('ok')
        self._login_admin()
        url = reverse('monitoring:api_location_geojson')
        response = self.client.get(url)
        data = response.data
        self.assertEqual(data['count'], 1)
        self.assertEqual(len(data['features']), 1)
        self.assertEqual(data['features'][0]['properties']['device_count'], 1)
        self.assertEqual(data['features'][0]['properties']['ok_count'], 1)
        self.assertEqual(data['features'][0]['properties']['problem_count'], 0)
        self.assertEqual(data['features'][0]['properties']['critical_count'], 0)
        self.assertEqual(data['features'][0]['properties']['unknown_count'], 0)

    def test_api_location_device_list(self):
        device_location = self._create_object_location()
        device_location.device.monitoring.update_status('ok')
        location = device_location.location
        self._login_admin()
        url = reverse('monitoring:api_location_device_list', args=[location.pk])
        response = self.client.get(url)
        data = response.data
        self.assertEqual(data['count'], 1)
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(device_location.device.id))
        self.assertIn('monitoring', data['results'][0])
        self.assertIsInstance(data['results'][0]['monitoring'], dict)
        self.assertDictEqual(
            data['results'][0]['monitoring'], {'status': 'ok', 'status_label': 'ok'}
        )

    def test_api_monitoring_nearby_device_list(self):
        admin = self._create_admin()
        self.client.force_login(admin)
        org1 = self._get_org()
        org2 = self._create_org(name='org2', slug='org2')
        device_without_location = self._create_device(organization=org1)
        org1_device1 = self._create_device(
            mac_address='11:22:33:44:55:66',
            name='device1',
            organization=org1,
            model='TP-Link Archer C20',
        )
        org1_device2 = self._create_device(
            mac_address='11:22:33:44:55:67',
            name='device2',
            organization=org1,
            model='TP-Link Archer C50',
        )
        org2_device1 = self._create_device(
            mac_address='11:22:33:44:55:68',
            name='device3',
            organization=org2,
            model='TP-Link Archer C60',
        )
        org2_device1.monitoring.status = 'ok'
        org2_device1.monitoring.save()

        self._create_object_location(
            content_object=org1_device1,
            location=Location.objects.create(
                organization=org1,
                name='location1',
                address='Uttrakhand',
                geometry='SRID=4326;POINT (79.0676 30.7333)',
                type='outdoor',
            ),
        )
        self._create_object_location(
            content_object=org1_device2,
            location=Location.objects.create(
                organization=org1,
                name='location2',
                address='Sunnyvale',
                geometry='SRID=4326;POINT (-122.03749 37.37187)',
                type='outdoor',
            ),
        )
        self._create_object_location(
            content_object=org2_device1,
            location=Location.objects.create(
                organization=org2,
                name='location3',
                address='Brussels',
                geometry='SRID=4326;POINT (4.406264339 50.89806471)',
                type='outdoor',
            ),
        )
        with self.subTest('Test DeviceLocation does not exist'):
            response = self.client.get(
                reverse(
                    'monitoring:api_monitoring_nearby_device_list',
                    args=[device_without_location.id],
                )
            )
            self.assertEqual(response.status_code, 404)

        path = reverse(
            'monitoring:api_monitoring_nearby_device_list', args=[org1_device1.id]
        )
        with self.subTest('Test device list'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 2)
            self.assertIn('distance', response.data['results'][0])
            self.assertIsInstance(response.data['results'][0]['distance'], float)
            self.assertIn('monitoring_status', response.data['results'][0])
            self.assertIn('monitoring_data', response.data['results'][0])

        with self.subTest('Test filtering by model'):
            response = self.client.get(path, data={'model': 'TP-Link Archer C50'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 1)
            self.assertEqual(response.data['results'][0]['model'], 'TP-Link Archer C50')
            # Test filtering with multiple models
            response = self.client.get(
                path, data={'model': 'TP-Link Archer C50|TP-Link Archer C60'}
            )
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 2)

        with self.subTest('Test filtering by monitoring status'):
            response = self.client.get(path, data={'monitoring__status': 'ok'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 1)
            self.assertEqual(response.data['results'][0]['monitoring_status'], 'ok')

        with self.subTest('Test filtering by distance'):
            response = self.client.get(path, data={'distance__lte': '6373400'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 1)
            self.assertEqual(response.data['results'][0]['distance'], 6373400)

        with self.subTest('Test pagination'):
            response = self.client.get(path, data={'page_size': '1'})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 2)
            self.assertEqual(len(response.data['results']), 1)
            self.assertEqual(response.data['previous'], None)
            self.assertNotEqual(response.data['next'], None)

        user = self._create_org_user(is_admin=True, organization=org1).user
        user.groups.add(Group.objects.get(name='Administrator'))
        self.client.force_login(user)
        with self.subTest('Test multi-tenancy'):
            response = self.client.get(path)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 1)
            self.assertEqual(response.data['results'][0]['id'], str(org1_device2.id))
            self.assertNotEqual(response.data['results'][0]['id'], str(org2_device1.id))

        self.client.logout()
        with self.subTest('Test device key authentication'):
            response = self.client.get(path, data={'key': org1_device1.key})
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 2)
            # Test with inccorect device key
            response = self.client.get(path, data={'key': 'incorrect'})
            self.assertEqual(response.status_code, 404)

    @capture_any_output()
    def test_bearer_authentication(self):
        user = self._create_admin()
        token = Token.objects.create(user=user).key
        device_location = self._create_object_location()
        location = device_location.location

        with self.subTest('Test MonitoringGeoJsonLocationList'):
            response = self.client.get(
                reverse('monitoring:api_location_geojson'),
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Bearer {token}',
            )
            self.assertEqual(response.status_code, 200)

        with self.subTest('Test GeoJsonLocationListView'):
            response = self.client.get(
                reverse('monitoring:api_location_device_list', args=[location.id]),
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Bearer {token}',
            )
            self.assertEqual(response.status_code, 200)

        with self.subTest('Test MonitoringNearbyDeviceList'):
            response = self.client.get(
                reverse(
                    'monitoring:api_monitoring_nearby_device_list',
                    args=[device_location.content_object_id],
                ),
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Bearer {token}',
            )
            self.assertEqual(response.status_code, 200)


class TestWifiSessionApi(
    AuthenticationMixin,
    TestMultitenantAdminMixin,
    TestWifiClientSessionMixin,
    CreateDeviceGroupMixin,
    DeviceMonitoringTestCase,
):
    def _login_admin(self):
        self._login()

    def _serialize_wifi_session(self, wifi_session, many=False, list_single=False):
        if many:
            serializer = WifiSessionSerializer(wifi_session, many=many)
            return serializer.data
        if list_single:
            serializer = WifiSessionSerializer(wifi_session)
            return serializer.data
        serializer = WifiSessionSerializer()
        return dict(serializer.to_representation(wifi_session))

    def _create_wifi_session_multi_env(self):
        org1 = self._get_org()
        dg1 = self._create_device_group(organization=org1)
        d1 = self._create_device(
            mac_address='00:11:22:33:44:66', group=dg1, organization=org1
        )
        ws1 = self._create_wifi_session(device=d1)
        ws1.wifi_client
        org2 = self._create_org(name='test-org-2')
        dg2 = self._create_device_group(organization=org2)
        d2 = self._create_device(
            mac_address='00:11:22:33:44:77', group=dg2, organization=org2
        )
        wc2 = self._create_wifi_client(mac_address='22:33:44:55:66:88')
        ws2 = self._create_wifi_session(device=d2, wifi_client=wc2)
        return org1, org2, d1, d2, dg1, dg2, ws1, ws2

    def _assert_wifi_session_list_filters(self, query_num, ws, filter_params={}):
        url = reverse('monitoring:api_wifi_session_list')
        with self.assertNumQueries(query_num):
            response = self.client.get(url, filter_params)
        self.assertEqual(response.status_code, 200)
        serializer_dict = self._serialize_wifi_session(ws, list_single=True)
        self.assertEqual(len(response.data['results']), 1)
        self.assertEqual(response.data['results'][0], serializer_dict)

    def test_wifi_session_list_unauthorized(self):
        device = self._create_device()
        wifi_session = self._create_wifi_session(device=device)
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('monitoring:api_wifi_session_detail', args=[wifi_session.id])
            )
        self.assertEqual(response.status_code, 401)
        response = self.client.get(reverse('monitoring:api_wifi_session_list'))
        self.assertEqual(response.status_code, 401)

    def test_wifi_session_detail_unauthorized(self):
        wifi_session = self._create_wifi_session()
        url = reverse('monitoring:api_wifi_session_detail', args=[wifi_session.id])
        with self.assertNumQueries(0):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 401)

    def test_wifi_session_detail_404(self):
        wifi_session_id = uuid4()
        self._login_admin()
        url = reverse('monitoring:api_wifi_session_detail', args=[wifi_session_id])
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_wifi_session_detail_get(self):
        device = self._create_device()
        wifi_session = self._create_wifi_session(device=device)
        wifi_session.wifi_client
        self._login_admin()
        url = reverse('monitoring:api_wifi_session_detail', args=[wifi_session.id])
        with self.assertNumQueries(2):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        serializer_dict = self._serialize_wifi_session(wifi_session)
        self.assertEqual(response.data, serializer_dict)

    def test_wifisession_list_get(self):
        self._create_wifi_session_multi_env()
        self._login_admin()
        url = reverse('monitoring:api_wifi_session_list')
        with self.assertNumQueries(3):
            response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        wifi_session_qs = WifiSession.objects.order_by('-start_time')
        serializer_list = self._serialize_wifi_session(wifi_session_qs, many=True)
        self.assertEqual(response.data['results'], serializer_list)

    def test_wifisession_list_filters(self):
        org1, org2, d1, d2, dg1, dg2, ws1, ws2 = self._create_wifi_session_multi_env()
        self._login_admin()

        with self.subTest('Test filtering using organization id'):
            self._assert_wifi_session_list_filters(
                4, ws1, {'device__organization': org1.id}
            )
            self._assert_wifi_session_list_filters(
                4, ws2, {'device__organization': org2.id}
            )

        with self.subTest('Test filtering using device id'):
            self._assert_wifi_session_list_filters(4, ws1, {'device': d1.id})
            self._assert_wifi_session_list_filters(4, ws2, {'device': d2.id})

        with self.subTest('Test filtering using device group id'):
            self._assert_wifi_session_list_filters(4, ws1, {'device__group': dg1.id})
            self._assert_wifi_session_list_filters(4, ws2, {'device__group': dg2.id})

        with self.subTest('Test filtering using start time'):
            filter_time = ws2.start_time
            url = reverse('monitoring:api_wifi_session_list')
            wifi_session_qs = WifiSession.objects.order_by('-start_time')
            response = self.client.get(url, {'start_time': filter_time})
            self.assertEqual(response.data['count'], 1)
            # Make sure the correct WiFi session is returned in the response
            serializer_dict = self._serialize_wifi_session(ws2, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)
            response = self.client.get(url, {'start_time__gt': filter_time})
            self.assertEqual(response.data['count'], 0)
            response = self.client.get(url, {'start_time__gte': filter_time})
            self.assertEqual(response.data['count'], 1)
            serializer_dict = self._serialize_wifi_session(ws2, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)
            response = self.client.get(url, {'start_time__lt': filter_time})
            serializer_dict = self._serialize_wifi_session(ws1, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)
            response = self.client.get(url, {'start_time__lte': filter_time})
            self.assertEqual(response.data['count'], 2)
            serializer_dict = self._serialize_wifi_session(wifi_session_qs, many=True)
            self.assertEqual(response.data['results'], serializer_dict)

        with self.subTest('Test filtering using stop time'):
            ws2.stop_time = timezone.now()
            ws2.save()
            ws1.stop_time = timezone.now()
            ws1.save()
            filter_time = ws2.stop_time
            url = reverse('monitoring:api_wifi_session_list')
            wifi_session_qs = WifiSession.objects.order_by('-start_time')
            response = self.client.get(url, {'stop_time': filter_time})
            self.assertEqual(response.data['count'], 1)
            serializer_dict = self._serialize_wifi_session(ws2, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)
            response = self.client.get(url, {'stop_time__gt': filter_time})
            self.assertEqual(response.data['count'], 1)
            serializer_dict = self._serialize_wifi_session(ws1, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)
            response = self.client.get(url, {'stop_time__gte': filter_time})
            self.assertEqual(response.data['count'], 2)
            serializer_dict = self._serialize_wifi_session(wifi_session_qs, many=True)
            self.assertEqual(response.data['results'], serializer_dict)
            response = self.client.get(url, {'stop_time__lt': filter_time})
            self.assertEqual(response.data['count'], 0)
            response = self.client.get(url, {'stop_time__lte': filter_time})
            self.assertEqual(response.data['count'], 1)
            serializer_dict = self._serialize_wifi_session(ws2, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)

    def test_wifi_session_list_detail_multitenancy(self):
        org1, org2, d1, d2, dg1, dg2, ws1, ws2 = self._create_wifi_session_multi_env()
        org1_manager = self._create_operator(
            organizations=[org1],
            username='org1_manager',
            email='orgmanager@test.com',
        )
        org1_member = self._create_operator(
            username='org1_member', email='orgmember@test.com'
        )
        org1_admin = self._create_operator(
            username='org_admin', email='org_admin@test.com', is_superuser=True
        )

        with self.subTest('Test wifi session list org manager'):
            self.client.force_login(org1_manager)
            url = reverse('monitoring:api_wifi_session_list')
            with self.assertNumQueries(5):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.data['count'], 1)
            serializer_dict = self._serialize_wifi_session(ws1, list_single=True)
            self.assertEqual(response.data['results'][0], serializer_dict)

        with self.subTest('Test wifi session detail org manager'):
            self.client.force_login(org1_manager)
            url = reverse('monitoring:api_wifi_session_detail', args=[ws1.id])
            with self.assertNumQueries(4):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 200)
            serializer_dict = self._serialize_wifi_session(ws1)
            self.assertEqual(response.data, serializer_dict)
            # Make sure the WiFi session belonging to org2 is not accessible
            url = reverse('monitoring:api_wifi_session_detail', args=[ws2.pk])
            with self.assertNumQueries(4):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 404)

        with self.subTest('Test wifi session list org member (403 forbidden)'):
            self.client.force_login(org1_member)
            url = reverse('monitoring:api_wifi_session_list')
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(1):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 403)
            self.assertIn(err, response.json()['detail'])

        with self.subTest('Test wifi session detail org member (403 forbidden)'):
            self.client.force_login(org1_member)
            url = reverse('monitoring:api_wifi_session_detail', args=[ws1.pk])
            err = 'User is not a manager of the organization'
            with self.assertNumQueries(1):
                response = self.client.get(url)
            self.assertEqual(response.status_code, 403)
            self.assertIn(err, response.json()['detail'])
            url = reverse('monitoring:api_wifi_session_detail', args=[ws2.pk])
            with self.assertNumQueries(1):
                r = self.client.get(url)
            self.assertEqual(response.status_code, 403)
            self.assertIn(err, response.json()['detail'])

        with self.subTest('Test upgrade operation list org admin'):
            # Superuser (org1_admin) can view
            # wifi sessions of both organizations
            self.client.force_login(org1_admin)
            wifi_session_qs = WifiSession.objects.order_by('-start_time')
            url = reverse('monitoring:api_wifi_session_list')
            with self.assertNumQueries(3):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_wifi_session(wifi_session_qs, many=True)
            self.assertEqual(r.data['results'], serializer_list)
            url = reverse('monitoring:api_wifi_session_list')
            with self.assertNumQueries(3):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_list = self._serialize_wifi_session(wifi_session_qs, many=True)
            self.assertEqual(r.data['results'], serializer_list)
            # Ensure similar results for wifi_session detail
            url = reverse('monitoring:api_wifi_session_detail', args=[ws1.pk])
            with self.assertNumQueries(2):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_dict = self._serialize_wifi_session(ws1)
            self.assertEqual(r.data, serializer_dict)
            url = reverse('monitoring:api_wifi_session_detail', args=[ws2.pk])
            with self.assertNumQueries(2):
                r = self.client.get(url)
            self.assertEqual(r.status_code, 200)
            serializer_dict = self._serialize_wifi_session(ws2)
            self.assertEqual(r.data, serializer_dict)



from datetime import timedelta

from django.utils import timezone
from freezegun import freeze_time
from swapper import load_model

from ...device.tests import (
    DeviceMonitoringTestCase,
    DeviceMonitoringTransactionTestcase,
)

Metric = load_model('monitoring', 'Metric')
Notification = load_model('openwisp_notifications', 'Notification')
Device = load_model('config', 'Device')
Config = load_model('config', 'Config')
OrganizationUser = load_model('openwisp_users', 'OrganizationUser')

notification_queryset = Notification.objects.order_by('timestamp')
start_time = timezone.now()
ten_minutes_ago = start_time - timedelta(minutes=10)
ten_minutes_after = start_time + timedelta(minutes=10)


class TestMonitoringNotifications(DeviceMonitoringTestCase):
    device_model = Device
    config_model = Config

    def test_general_check_threshold_crossed_immediate(self):
        admin = self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )

        with self.subTest('Test notification for metric exceeding alert settings'):
            m.write(99)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'warning')

        with self.subTest('Test no double alarm for metric exceeding alert settings'):
            m.write(95)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)

        with self.subTest('Test notification for metric falling behind alert settings'):
            m.write(60)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)
            n = notification_queryset.last()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'info')

        with self.subTest(
            'Test no double alarm for metric falling behind alert settings'
        ):
            m.write(40)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)

    def test_general_check_threshold_crossed_deferred(self):
        admin = self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        with freeze_time(start_time):
            m.write(99)
        with freeze_time(ten_minutes_after):
            m.write(99)
        m.refresh_from_db()
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        n = notification_queryset.first()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, m)
        self.assertEqual(n.action_object, m.alertsettings)
        self.assertEqual(n.level, 'warning')

    def test_general_check_threshold_deferred_not_crossed(self):
        self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        m.write(99)
        m.refresh_from_db(fields=['is_healthy', 'is_healthy_tolerant'])
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)

    def test_resources_metric_threshold_deferred_not_crossed(self):
        # Verify that newly created metrics within threshold limit don't send any notifications
        self._create_admin()
        self.create_test_data()
        self.assertEqual(Notification.objects.count(), 0)

    def test_general_check_threshold_crossed_for_long_time(self):
        """Test threshold remains crossed for a long time.

        This is the most common scenario: incoming metrics will always be
        stored with the current timestamp, which means the system must be
        able to look back in previous measurements to see if the
        AlertSettings has been crossed for long enough.
        """
        admin = self._create_admin()
        m = self._create_general_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=5
        )

        with self.subTest("Test no notification is generated for healthy status"):
            with freeze_time(ten_minutes_ago):
                m.write(89)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)

        # When using UDP for writing data to the timeseries, reading the
        # metric from the database provides a time delay that allows the
        # timeseries database to process the transaction.
        self._read_metric(m)
        with self.subTest("Test no notification is generated when check=False"):
            with freeze_time(ten_minutes_ago):
                m.write(91, check=False)
            self.assertEqual(Notification.objects.count(), 0)

        self._read_metric(m)
        with self.subTest('Test notification for metric with current timestamp'):
            m.write(92)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'warning')

        self._read_metric(m)
        with self.subTest('Test no recovery notification yet (tolerance not passed)'):
            m.write(50)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)

        self._read_metric(m)
        with self.subTest('Tolerance still not passed, not expecting a recovery yet'):
            with freeze_time(start_time + timedelta(minutes=2)):
                m.write(51)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)

        self._read_metric(m)
        with self.subTest('Test recovery notification after tolerance is passed'):
            with freeze_time(ten_minutes_after):
                m.write(50)
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)
            n = notification_queryset.last()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'info')

    def test_object_check_threshold_crossed_immediate(self):
        admin = self._create_admin()
        om = self._create_object_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )

        with self.subTest(
            "Test notification for object metric exceeding alert settings"
        ):
            om.write(99)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, False)
            self.assertEqual(om.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, om)
            self.assertEqual(n.action_object, alert_s)
            self.assertEqual(n.target, om.content_object)
            self.assertEqual(n.level, 'warning')

        with self.subTest(
            "Test no double alarm for object metric exceeding alert settings"
        ):
            om.write(95)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, False)
            self.assertEqual(om.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)

        with self.subTest(
            "Test notification for object metric falling behind alert settings"
        ):
            om.write(60)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, True)
            self.assertEqual(om.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)
            n = notification_queryset.last()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, om)
            self.assertEqual(n.action_object, alert_s)
            self.assertEqual(n.target, om.content_object)
            self.assertEqual(n.level, 'info')

        with self.subTest(
            "Test no double alarm for object metric falling behind alert settings"
        ):
            om.write(40)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, True)
            self.assertEqual(om.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 2)

    def test_object_check_threshold_crossed_deferred(self):
        admin = self._create_admin()
        om = self._create_object_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        with freeze_time(start_time):
            om.write(99)
        with freeze_time(ten_minutes_after):
            om.write(99)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        n = notification_queryset.first()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, om)
        self.assertEqual(n.action_object, alert_s)
        self.assertEqual(n.target, om.content_object)
        self.assertEqual(n.level, 'warning')

    def test_object_check_threshold_deferred_not_crossed(self):
        self._create_admin()
        om = self._create_object_metric(name='load')
        self._create_alert_settings(
            metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        om.write(99)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)

    def test_object_check_threshold_crossed_for_long_time(self):
        admin = self._create_admin()
        om = self._create_object_metric(name='load')
        alert_s = self._create_alert_settings(
            metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=1
        )
        with freeze_time(ten_minutes_ago):
            self._write_metric(om, 89)
        self.assertEqual(Notification.objects.count(), 0)
        with freeze_time(ten_minutes_ago):
            self._write_metric(om, 91, check=False)
        self.assertEqual(Notification.objects.count(), 0)
        self._write_metric(om, 92)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        n = notification_queryset.first()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, om)
        self.assertEqual(n.action_object, alert_s)
        self.assertEqual(n.target, om.content_object)
        self.assertEqual(n.level, 'warning')
        # ensure double alarm not sent
        self._write_metric(om, 95)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        # value back to normal but tolerance not passed yet
        self._write_metric(om, 60)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, True)
        self.assertEqual(om.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        # tolerance passed
        with freeze_time(ten_minutes_after):
            self._write_metric(om, 60)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, True)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 2)
        n = notification_queryset.last()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, om)
        self.assertEqual(n.action_object, alert_s)
        self.assertEqual(n.target, om.content_object)
        self.assertEqual(n.level, 'info')
        # ensure double alarm not sent
        with freeze_time(ten_minutes_after + timedelta(minutes=5)):
            self._write_metric(om, 40)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, True)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 2)

    @freeze_time(start_time)
    def test_object_check_threshold_crossed_historical_data(self):
        """
        Do not evaluate threshold crossed for historical data
        """
        self._create_admin()
        om = self._create_object_metric(name="load")
        self._create_alert_settings(
            metric=om, custom_operator=">", custom_threshold=90, custom_tolerance=1
        )

        # We need to write with "time" argument instead of freeze_time to
        # simulate historical data
        self._write_metric(om, 99, time=start_time - timedelta(minutes=60))
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, True)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)

        self._write_metric(om, 99, time=start_time - timedelta(minutes=10))
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, True)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)

        # Writing real-time data should enforce the threshold check
        self._write_metric(om, 99, time=start_time)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)

    def test_flapping_metric_with_tolerance(self):
        self._create_admin()
        om = self._create_object_metric(name='ping')
        self._create_alert_settings(
            metric=om, custom_operator='<', custom_threshold=1, custom_tolerance=10
        )
        with freeze_time(start_time - timedelta(minutes=25)):
            om.write(1)
            self.assertEqual(Notification.objects.count(), 0)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, True)
            self.assertEqual(om.is_healthy_tolerant, True)
        with freeze_time(start_time - timedelta(minutes=20)):
            om.write(0)
            self.assertEqual(Notification.objects.count(), 0)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, False)
            self.assertEqual(om.is_healthy_tolerant, True)
        with freeze_time(start_time - timedelta(minutes=15)):
            om.write(1)
            self.assertEqual(Notification.objects.count(), 0)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, True)
            self.assertEqual(om.is_healthy_tolerant, True)
        with freeze_time(start_time - timedelta(minutes=10)):
            om.write(0)
            self.assertEqual(Notification.objects.count(), 0)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, False)
            self.assertEqual(om.is_healthy_tolerant, True)
        with freeze_time(start_time - timedelta(minutes=5)):
            om.write(0)
            self.assertEqual(Notification.objects.count(), 0)
            om.refresh_from_db()
            self.assertEqual(om.is_healthy, False)
            self.assertEqual(om.is_healthy_tolerant, True)
        om.write(0)
        self.assertEqual(Notification.objects.count(), 1)
        om.refresh_from_db()
        self.assertEqual(om.is_healthy, False)
        self.assertEqual(om.is_healthy_tolerant, False)

    def test_notification_types(self):
        self._create_admin()
        m = self._create_object_metric(name='load')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=0
        )
        exp_message = (
            '{n.actor.name} for device '
            '<a href="https://example.com/admin/openwisp_users/user/{n.target.id}/change/">tester</a>'
            ' {n.verb}.'
        )
        with self.subTest("Test notification for 'alert settings crossed'"):
            m.write(99)
            n = notification_queryset.first()
            self.assertEqual(n.level, 'warning')
            self.assertEqual(n.verb, 'crossed the threshold')
            self.assertEqual(
                n.email_subject, f'[example.com] PROBLEM: {n.actor.name} {n.target}'
            )
            self.assertIn(exp_message.format(n=n), n.message)

        with self.subTest("Test notification for 'under alert settings'"):
            m.write(80)
            n = notification_queryset.last()
            self.assertEqual(n.level, 'info')
            self.assertEqual(n.verb, 'returned within the threshold')
            self.assertEqual(
                n.email_subject, f'[example.com] RECOVERY: {n.actor.name} {n.target}'
            )
            self.assertIn(exp_message.format(n=n), n.message)

    def test_alerts_disabled(self):
        self._create_admin()
        d = self._create_device(organization=self._get_org())
        m = self._create_general_metric(name='load', content_object=d)
        self._create_alert_settings(
            metric=m,
            custom_operator='>',
            custom_threshold=90,
            custom_tolerance=1,
            is_active=False,
        )
        with freeze_time(start_time):
            m.write(99)
        with freeze_time(ten_minutes_after):
            m.write(99)
        m.refresh_from_db()
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, False)
        d.refresh_from_db()
        self.assertEqual(d.monitoring.status, 'problem')
        self.assertEqual(Notification.objects.count(), 0)

    def test_alert_field(self):
        admin = self._create_admin()

        def _create_alert_field_test_env():
            m = self._create_general_metric(configuration='test_alert_field')
            self._create_alert_settings(
                metric=m, custom_operator='>', custom_threshold=30, custom_tolerance=0
            )
            return m

        with self.subTest('Test notification for metric without related field'):
            m = _create_alert_field_test_env()
            with self.assertRaises(ValueError) as err:
                m.write(1)
            self.assertEqual(
                str(err.exception),
                'write() missing keyword argument: "extra_values" required for alert on related field',
            )
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)

        with self.subTest('Test notification for metric on different related field'):
            m = _create_alert_field_test_env()
            with self.assertRaises(ValueError) as err:
                m.write(10, extra_values={'test_related_3': 40})
            self.assertEqual(
                str(err.exception),
                '"test_related_3" is not defined for alert_field in metric configuration',
            )
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)

        with self.subTest('Test notification for metric with multiple related fields'):
            m = _create_alert_field_test_env()
            m.write(10, extra_values={'test_related_2': 40, 'test_related_3': 20})
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'warning')
        Notification.objects.all().delete()

        with self.subTest(
            'Test notification for metric exceeding related field alert settings'
        ):
            m = _create_alert_field_test_env()
            m.write(10, extra_values={'test_related_2': 40})
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self.assertEqual(n.recipient, admin)
            self.assertEqual(n.actor, m)
            self.assertEqual(n.action_object, m.alertsettings)
            self.assertEqual(n.level, 'warning')
        Notification.objects.all().delete()

        with self.subTest(
            'Test notification for metric falling behind related field alert settings'
        ):
            m = _create_alert_field_test_env()
            m.write(30, extra_values={'test_related_2': 25})
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, True)
            self.assertEqual(m.is_healthy_tolerant, True)
            self.assertEqual(Notification.objects.count(), 0)

        with self.subTest(
            'Test no double alarm for metric exceeding related field alert settings'
        ):
            m = _create_alert_field_test_env()
            m.write(20, extra_values={'test_related_2': 35})
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            Notification.objects.all().delete()
            # check for double alarm
            m.write(20, extra_values={'test_related_2': 40})
            m.refresh_from_db()
            self.assertEqual(m.is_healthy, False)
            self.assertEqual(m.is_healthy_tolerant, False)
            self.assertEqual(Notification.objects.count(), 0)

    def test_general_check_threshold_with_alert_field_crossed_deferred(self):
        admin = self._create_admin()
        m = self._create_general_metric(configuration='test_alert_field')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=30, custom_tolerance=1
        )
        with freeze_time(start_time):
            m.write(10, extra_values={"test_related_2": 35})
        with freeze_time(ten_minutes_after):
            m.write(10, extra_values={"test_related_2": 35})
        m.refresh_from_db()
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, False)
        self.assertEqual(Notification.objects.count(), 1)
        n = notification_queryset.first()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, m)
        self.assertEqual(n.action_object, m.alertsettings)
        self.assertEqual(n.level, 'warning')

    def test_general_check_threshold_with_alert_field_deferred_not_crossed(self):
        self._create_admin()
        m = self._create_general_metric(configuration='test_alert_field')
        self._create_alert_settings(
            metric=m, custom_operator='>', custom_threshold=30, custom_tolerance=1
        )
        m.write(10, extra_values={'test_related_2': 32})
        m.refresh_from_db(fields=['is_healthy', 'is_healthy_tolerant'])
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)
        m.write(20, extra_values={'test_related_2': 35})
        self.assertEqual(m.is_healthy, False)
        self.assertEqual(m.is_healthy_tolerant, True)
        self.assertEqual(Notification.objects.count(), 0)


class TestTransactionMonitoringNotifications(DeviceMonitoringTransactionTestcase):
    device_model = Device
    config_model = Config

    def _check_notification_parameters(self, notification, recepient, metric, target):
        self.assertEqual(notification.recipient, recepient)
        self.assertEqual(notification.actor, metric)
        self.assertEqual(notification.target, target)
        self.assertEqual(notification.action_object, metric.alertsettings)
        self.assertEqual(notification.level, 'warning')
        self.assertEqual(notification.verb, 'is not reachable')

    def test_cpu_metric_threshold_crossed(self):
        admin = self._create_admin()
        org = self._create_org()
        device = self._create_device(organization=org)
        # creates metric and alert settings
        data = self._data()
        data['resources']['load'] = [0.99, 0.99, 0.99]
        response = self._post_data(device.id, device.key, data)
        self.assertEqual(response.status_code, 200)
        # retrieve created metric
        metric = Metric.objects.get(name='CPU usage')
        # simplify test by setting tolerance to 0
        metric.alertsettings.custom_tolerance = 0
        metric.alertsettings.save()
        # trigger alert
        metric.write(99.0)
        self.assertEqual(Notification.objects.count(), 1)
        n = Notification.objects.first()
        self.assertEqual(n.recipient, admin)
        self.assertEqual(n.actor, metric)
        self.assertEqual(n.action_object, metric.alertsettings)
        self.assertEqual(n.level, 'warning')

    def test_multiple_notifications(self):
        testorg = self._create_org()
        admin = self._create_admin()
        staff = self._create_user(
            username='staff', email='staff@staff.com', password='staff', is_staff=True
        )
        self._create_user(
            username='staff-lone',
            email='staff-lone@staff.com',
            password='staff',
            is_staff=True,
        )
        user = self._create_user(is_staff=False)
        OrganizationUser.objects.create(user=user, organization=testorg, is_admin=True)
        OrganizationUser.objects.create(user=staff, organization=testorg, is_admin=True)
        self.assertIsNotNone(staff.notificationsetting_set.filter(organization=testorg))

        with self.subTest('Test general metric multiple notifications'):
            m = self._create_general_metric(name='load')
            alert_s = self._create_alert_settings(
                metric=m, custom_operator='>', custom_threshold=90, custom_tolerance=1
            )
            m._notify_users(notification_type='ping_problem', alert_settings=alert_s)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self._check_notification_parameters(n, admin, m, None)
            self.assertIn(
                'The device <a href="#">None</a> is not reachable.', n.message
            )
        Notification.objects.all().delete()

        with self.subTest('Test object metric multiple notifications'):
            d = self._create_device(organization=testorg)
            om = self._create_object_metric(name='load', content_object=d)
            alert_s = self._create_alert_settings(
                metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=1
            )
            self.assertEqual(Notification.objects.count(), 0)
            om._notify_users(notification_type='ping_problem', alert_settings=alert_s)
            self.assertEqual(Notification.objects.count(), 2)
            n = notification_queryset.first()
            self._check_notification_parameters(n, admin, om, d)
            self.assertIn('is not reachable.', n.message)
            n = notification_queryset.last()
            self._check_notification_parameters(n, staff, om, d)
        Notification.objects.all().delete()

        with self.subTest('Test object metric multiple notifications no org'):
            om = self._create_object_metric(name='logins', content_object=user)
            alert_s = self._create_alert_settings(
                metric=om, custom_operator='>', custom_threshold=90, custom_tolerance=0
            )
            self.assertEqual(Notification.objects.count(), 0)
            om._notify_users(notification_type='ping_problem', alert_settings=alert_s)
            self.assertEqual(Notification.objects.count(), 1)
            n = notification_queryset.first()
            self._check_notification_parameters(n, admin, om, user)



from django.contrib.contenttypes.fields import GenericRelation
from swapper import get_model_name, load_model, swappable_setting

from .base.models import (
    AbstractDeviceData,
    AbstractDeviceMonitoring,
    AbstractWifiClient,
    AbstractWifiSession,
)

BaseDevice = load_model('config', 'Device', require_ready=False)


class DeviceData(AbstractDeviceData, BaseDevice):
    checks = GenericRelation(get_model_name('check', 'Check'))
    metrics = GenericRelation(get_model_name('monitoring', 'Metric'))

    class Meta:
        proxy = True
        swappable = swappable_setting('device_monitoring', 'DeviceData')


class DeviceMonitoring(AbstractDeviceMonitoring):
    class Meta(AbstractDeviceMonitoring.Meta):
        abstract = False
        swappable = swappable_setting('device_monitoring', 'DeviceMonitoring')


class WifiClient(AbstractWifiClient):
    class Meta(AbstractWifiClient.Meta):
        abstract = False
        swappable = swappable_setting('device_monitoring', 'WifiClient')


class WifiSession(AbstractWifiSession):
    class Meta(AbstractWifiSession.Meta):
        abstract = False
        swappable = swappable_setting('device_monitoring', 'WifiSession')



# Generated by Django 3.0.3 on 2020-07-30 15:54

from django.db import migrations, models

from ..settings import CHECK_CLASSES


class Migration(migrations.Migration):
    dependencies = [('check', '0003_create_ping')]

    operations = [
        migrations.RenameField(
            model_name='check', old_name='active', new_name='is_active'
        ),
        migrations.AlterField(
            model_name='check',
            name='is_active',
            field=models.BooleanField(
                db_index=True,
                default=True,
                help_text='whether the check should be run, related metrics collected and alerts sent',
                verbose_name='active',
            ),
        ),
        migrations.AlterField(
            model_name='check',
            name='check',
            field=models.CharField(
                choices=CHECK_CLASSES,
                db_index=True,
                max_length=128,
                verbose_name='check type',
            ),
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-20 09:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('monitoring', '0006_migrate_influxdb_structure'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metric',
            name='object_id',
            field=models.CharField(blank=True, db_index=True, max_length=36, null=True),
        ),
    ]



from django.dispatch import Signal

threshold_crossed = Signal()
threshold_crossed.__doc__ = """
Providing arguments: ['metric', 'alert_settings', 'target', 'first_time']
"""
pre_metric_write = Signal()
pre_metric_write.__doc__ = """
Providing arguments: ['metric', 'values', 'time', 'current']
"""
post_metric_write = Signal()
post_metric_write.__doc__ = """
Providing arguments: ['metric', 'values', 'time', 'current']
"""



# Generated by Django 3.2.14 on 2022-08-12 07:50

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('check', '0007_create_checks'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='check',
            options={
                'permissions': (
                    ('add_check_inline', 'Can add check inline'),
                    ('change_check_inline', 'Can change check inline'),
                    ('delete_check_inline', 'Can delete check inline'),
                    ('view_check_inline', 'Can view check inline'),
                )
            },
        ),
    ]



from collections import OrderedDict

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from jsonfield import JSONField

from openwisp_monitoring.check import settings as app_settings
from openwisp_monitoring.check.tasks import (
    auto_create_config_check,
    auto_create_iperf3_check,
    auto_create_ping,
)
from openwisp_utils.base import TimeStampedEditableModel

from ...utils import transaction_on_commit


class AbstractCheck(TimeStampedEditableModel):
    name = models.CharField(max_length=64, db_index=True)
    is_active = models.BooleanField(
        _('active'),
        default=True,
        db_index=True,
        help_text=_(
            'whether the check should be run, related metrics collected and alerts sent'
        ),
    )
    description = models.TextField(blank=True, help_text=_('Notes'))
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, blank=True
    )
    object_id = models.CharField(max_length=36, db_index=True, blank=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    check_type = models.CharField(
        _('check type'),
        choices=app_settings.CHECK_CLASSES,
        db_index=True,
        max_length=128,
    )
    params = JSONField(
        _('parameters'),
        default=dict,
        blank=True,
        help_text=_('parameters needed to perform the check'),
        load_kwargs={'object_pairs_hook': OrderedDict},
        dump_kwargs={'indent': 4},
    )

    class Meta:
        abstract = True
        unique_together = ('name', 'object_id', 'content_type')

        permissions = (
            ('add_check_inline', 'Can add check inline'),
            ('change_check_inline', 'Can change check inline'),
            ('delete_check_inline', 'Can delete check inline'),
            ('view_check_inline', 'Can view check inline'),
        )

    def __str__(self):
        if not self.object_id or not self.content_type:
            return self.name
        obj = self.content_object
        model_name = obj.__class__.__name__
        return '{0} ({1}: {2})'.format(self.name, model_name, obj)

    def clean(self):
        self.check_instance.validate()

    def full_clean(self, *args, **kwargs):
        # The name of the check will be the same as the
        # 'check_type' chosen by the user when the
        # name field is empty (useful for CheckInline)
        if not self.name:
            self.name = self.get_check_type_display()
        return super().full_clean(*args, **kwargs)

    @cached_property
    def check_class(self):
        """Returns the check class."""
        return import_string(self.check_type)

    @cached_property
    def check_instance(self):
        """Returns the check class instance."""
        check_class = self.check_class
        return check_class(check=self, params=self.params)

    def perform_check(self, store=True):
        """Initializes check instance and calls the check method."""
        if (
            hasattr(self.content_object, 'is_deactivated')
            and self.content_object.is_deactivated()
        ) or (
            hasattr(self.content_object, 'organization_id')
            and self.content_object.organization.is_active is False
        ):
            return
        return self.check_instance.check(store=True)

    def perform_check_delayed(self, duration=0):
        from ..tasks import perform_check

        perform_check.apply_async(args=[self.id], countdown=duration)


def auto_ping_receiver(sender, instance, created, **kwargs):
    """Implements OPENWISP_MONITORING_AUTO_PING.

    The creation step is executed in the background.
    """
    # we need to skip this otherwise this task will be executed
    # every time the configuration is requested via checksum
    if not created:
        return
    transaction_on_commit(
        lambda: auto_create_ping.delay(
            model=sender.__name__.lower(),
            app_label=sender._meta.app_label,
            object_id=str(instance.pk),
        )
    )


def auto_config_check_receiver(sender, instance, created, **kwargs):
    """Implements OPENWISP_MONITORING_AUTO_DEVICE_CONFIG_CHECK.

    The creation step is executed in the background.
    """
    # we need to skip this otherwise this task will be executed
    # every time the configuration is requested via checksum
    if not created:
        return
    transaction_on_commit(
        lambda: auto_create_config_check.delay(
            model=sender.__name__.lower(),
            app_label=sender._meta.app_label,
            object_id=str(instance.pk),
        )
    )


def auto_iperf3_check_receiver(sender, instance, created, **kwargs):
    """Implements OPENWISP_MONITORING_AUTO_IPERF3.

    The creation step is executed in the background.
    """
    # we need to skip this otherwise this task will be executed
    # every time the configuration is requested via checksum
    if not created:
        return
    transaction_on_commit(
        lambda: auto_create_iperf3_check.delay(
            model=sender.__name__.lower(),
            app_label=sender._meta.app_label,
            object_id=str(instance.pk),
        )
    )



from cache_memoize import cache_memoize
from django.db.models import Q
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.permissions import IsAuthenticated
from rest_framework.views import APIView
from swapper import load_model

from openwisp_users.api.mixins import ProtectedAPIMixin

from ...settings import CACHE_TIMEOUT
from ...views import MonitoringApiViewMixin
from ..configuration import DEFAULT_DASHBOARD_TRAFFIC_CHART

Chart = load_model('monitoring', 'Chart')
Metric = load_model('monitoring', 'Metric')
Organization = load_model('openwisp_users', 'Organization')
Location = load_model('geo', 'Location')
FloorPlan = load_model('geo', 'FloorPlan')


class DashboardTimeseriesView(ProtectedAPIMixin, MonitoringApiViewMixin, APIView):
    """Multi-tenant view that returns general monitoring charts for the admin dashboard.

    Allows filtering with organization slugs.
    """

    permission_classes = [IsAuthenticated]

    def _get_organizations(self):
        query_orgs = self.request.query_params.get('organization_slug', '')
        if query_orgs:
            query_orgs = query_orgs.split(',')
            org_ids = list(
                Organization.objects.filter(slug__in=query_orgs).values_list(
                    'id', flat=True
                )
            )
            if len(org_ids) != len(query_orgs):
                # Request query_params contains slugs of non-existing organization
                raise NotFound
            if self.request.user.is_superuser or all(
                [str(id) in self.request.user.organizations_managed for id in org_ids]
            ):
                return org_ids
            # Request query_params contains slugs for organizations that user
            # doesn't manage
            raise PermissionDenied
        else:
            if self.request.user.is_superuser:
                return []
            elif self.request.user.organizations_managed:
                return self.request.user.organizations_managed
            else:
                # The user does not manage any organization
                raise PermissionDenied

    def _get_interface(self):
        orgs = self._get_organizations()
        interfaces = []
        if len(orgs) == 1:
            interfaces = DEFAULT_DASHBOARD_TRAFFIC_CHART.get(str(orgs[0]), [])
        if not interfaces:
            interfaces = DEFAULT_DASHBOARD_TRAFFIC_CHART.get('__all__')
        return interfaces

    def _get_location_ids(self):
        location_ids = self.request.query_params.get('location_id', '')
        if location_ids:
            location_ids = location_ids.split(',')
            location_query = Q(id__in=location_ids)
            if not self.request.user.is_superuser:
                location_query &= Q(organization__in=self._get_organizations())
            location_orgs = Location.objects.filter(location_query).values_list(
                'organization_id', flat=True
            )
            if len(location_ids) != len(location_orgs):
                # Request query_params contains slugs of non-existing organization
                raise NotFound
            if self.request.user.is_superuser or all(
                [
                    str(id) in self.request.user.organizations_managed
                    for id in location_orgs
                ]
            ):
                return location_ids
            # Request query_params contains ids of location from organizations that user
            # doesn't manage
            raise PermissionDenied
        else:
            if (
                self.request.user.is_superuser
                or self.request.user.organizations_managed
            ):
                # Don't filter by location by default
                return []
            else:
                # The user does not manage any organization
                raise PermissionDenied

    def _get_floorplan_ids(self):
        floorplan_ids = self.request.query_params.get('floorplan_id', '')
        if floorplan_ids:
            floorplan_ids = floorplan_ids.split(',')
            floorplan_query = Q(id__in=floorplan_ids)
            if not self.request.user.is_superuser:
                locations = self._get_location_ids()
                if locations:
                    floorplan_query &= Q(location__in=locations)
                else:
                    orgs = self._get_organizations()
                    floorplan_query &= Q(location__organization__in=orgs)
            floorplan_count = FloorPlan.objects.filter(floorplan_query).count()
            if len(floorplan_ids) != floorplan_count:
                # Request query_params contains slugs of non-existing organization
                raise NotFound
            if (
                self.request.user.is_superuser
                or self.request.user.organizations_managed
            ):
                return floorplan_ids
            # Request query_params contains ids of floorplan from organizations that user
            # doesn't manage
            raise PermissionDenied
        else:
            if (
                self.request.user.is_superuser
                or self.request.user.organizations_managed
            ):
                # Don't filter by floorplan by default
                return []
            else:
                # The user does not manage any organization
                raise PermissionDenied

    def _get_chart_additional_query_kwargs(self, chart):
        additional_params = {
            'organization_id': self._get_organizations(),
            'location_id': self._get_location_ids(),
            'floorplan_id': self._get_floorplan_ids(),
        }
        if chart.configuration == 'general_traffic':
            additional_params['ifname'] = self._get_interface()
        return {'additional_params': additional_params}

    @cache_memoize(
        CACHE_TIMEOUT,
        key_generator_callable=lambda *args, **kwargs: 'ow-monitoring-dashboard-charts',
    )
    def _get_charts(self, request, *args, **kwargs):
        return Chart.objects.filter(
            metric__object_id=None, metric__content_type=None
        ).select_related('metric')

    @classmethod
    def invalidate_cache(cls, instance, *args, **kwargs):
        if isinstance(instance, Chart):
            if (
                instance.metric.object_id is not None
                or instance.metric.content_type is not None
            ):
                return
        elif isinstance(instance, Metric):
            if instance.object_id is not None or instance.content_type is not None:
                return
        cls._get_charts.invalidate()

    def _get_user_managed_orgs(self, request):
        """Return list of dictionary containing organization name and slug in select2 compatible format."""
        orgs = []
        qs = Organization.objects.only('slug', 'name')
        if not request.user.is_superuser:
            if len(request.user.organizations_managed) > 1:
                qs = qs.filter(pk__in=request.user.organizations_managed)
            else:
                return orgs
        for org in qs.iterator():
            orgs.append({'id': org.slug, 'text': org.name})
        if len(orgs) < 2:
            # Handles scenarios for superuser when the project has only
            # one organization.
            return []
        return orgs

    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)
        if not request.GET.get('csv'):
            user_managed_orgs = self._get_user_managed_orgs(request)
            if user_managed_orgs:
                response.data['organizations'] = user_managed_orgs
        return response


dashboard_timeseries = DashboardTimeseriesView.as_view()



from django.conf import settings


def get_settings_value(option, default=None):
    return getattr(settings, f'OPENWISP_MONITORING_{option}', default)


AUTO_CHARTS = get_settings_value(
    'AUTO_CHARTS',
    (
        'traffic',
        'wifi_clients',
        'uptime',
        'packet_loss',
        'rtt',
        'memory',
        'cpu',
        'disk',
    ),
)

MONITORING_API_URLCONF = get_settings_value('API_URLCONF', None)
MONITORING_API_BASEURL = get_settings_value('API_BASEURL', None)
MONITORING_TIMESERIES_RETRY_OPTIONS = get_settings_value(
    'TIMESERIES_RETRY_OPTIONS', dict(max_retries=6, delay=2)
)
CACHE_TIMEOUT = get_settings_value(
    'CACHE_TIMEOUT',
    24 * 60 * 60,  # 24 hours in seconds
)
DEFAULT_CHART_TIME = get_settings_value('DEFAULT_CHART_TIME', '7d')



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_delete, post_save
from django.utils.translation import gettext_lazy as _
from swapper import get_model_name, load_model

from openwisp_utils.admin_theme.menu import register_menu_group

from ..db import timeseries_db
from .configuration import get_metric_configuration, register_metric_notifications


class MonitoringConfig(AppConfig):
    name = 'openwisp_monitoring.monitoring'
    label = 'monitoring'
    verbose_name = _('Network Monitoring')

    def ready(self):
        timeseries_db.create_database()
        setattr(settings, 'OPENWISP_ADMIN_SHOW_USERLINKS_BLOCK', True)
        metrics = get_metric_configuration()
        for metric_name, metric_config in metrics.items():
            register_metric_notifications(metric_name, metric_config)
        self.register_menu_groups()
        self.connect_metric_signals()

    def register_menu_groups(self):
        register_menu_group(
            position=80,
            config={
                'label': 'Monitoring',
                'items': {
                    1: {
                        'label': _('Metrics'),
                        'model': get_model_name('monitoring', 'Metric'),
                        'name': 'changelist',
                        'icon': 'ow-metrics',
                    },
                    2: {
                        'label': _('Checks'),
                        'model': get_model_name('check', 'Check'),
                        'name': 'changelist',
                        'icon': 'ow-monitoring-checks',
                    },
                },
                'icon': 'ow-monitoring',
            },
        )

    def connect_metric_signals(self):
        from .api.views import DashboardTimeseriesView

        Metric = load_model('monitoring', 'Metric')
        Chart = load_model('monitoring', 'Chart')
        AlertSettings = load_model('monitoring', 'AlertSettings')
        post_delete.connect(
            Metric.post_delete_receiver,
            sender=Metric,
            dispatch_uid='metric_post_delete_receiver',
        )
        post_save.connect(
            Metric.invalidate_cache,
            sender=Metric,
            dispatch_uid='post_save_invalidate_metric_cache',
        )
        post_delete.connect(
            Metric.invalidate_cache,
            sender=Metric,
            dispatch_uid='post_delete_invalidate_metric_cache',
        )
        post_save.connect(
            AlertSettings.invalidate_cache,
            sender=AlertSettings,
            dispatch_uid='post_save_invalidate_metric_cache',
        )
        post_delete.connect(
            AlertSettings.invalidate_cache,
            sender=AlertSettings,
            dispatch_uid='post_delete_invalidate_metric_cache',
        )
        post_save.connect(
            DashboardTimeseriesView.invalidate_cache,
            sender=Metric,
            dispatch_uid='post_save_dashboard_tsdb_view_invalidate_cache',
        )
        post_save.connect(
            DashboardTimeseriesView.invalidate_cache,
            sender=Chart,
            dispatch_uid='post_save_dashboard_tsdb_view_invalidate_cache',
        )
        post_delete.connect(
            DashboardTimeseriesView.invalidate_cache,
            sender=Metric,
            dispatch_uid='post_delete_dashboard_tsdb_view_invalidate_cache',
        )
        post_delete.connect(
            DashboardTimeseriesView.invalidate_cache,
            sender=Chart,
            dispatch_uid='post_delete_dashboard_tsdb_view_invalidate_cache',
        )



from ..settings import get_settings_value

ADDITIONAL_CHARTS = get_settings_value('CHARTS', {})
ADDITIONAL_METRICS = get_settings_value('METRICS', {})

RETRY_OPTIONS = get_settings_value(
    'WRITE_RETRY_OPTIONS',
    dict(
        max_retries=None, retry_backoff=True, retry_backoff_max=600, retry_jitter=True
    ),
)
ADDITIONAL_DASHBOARD_TRAFFIC_CHART = get_settings_value('DASHBOARD_TRAFFIC_CHART', {})
